"""Utils for vendor packages."""

import json
import os

from starkware.starknet.compiler.compile import compile_starknet_files


class JsonEncoder(json.JSONEncoder):
    """Encoder converting sets to lists."""

    def default(self, obj):
        if isinstance(obj, set):
            return list(obj)
        return json.JSONEncoder.default(self, obj)


def add_files_to_report(contracts_folder: str, report_dict):
    """Add zero coverage files to report."""
    for path, _, files in os.walk(contracts_folder):
        for name in files:
            f = os.path.join(path, name)
            if f not in report_dict and f.endswith(".cairo"):
                report_dict[f] = []


def process_file(file: str):
    """Get relative path."""
    cwd = os.getcwd()
    if file.startswith(cwd):
        return file[len(cwd) + 1 :]
    return file


def get_file_statements(files, cairo_path=None):
    """Get the statements from the filename."""
    if cairo_path is None:
        cairo_path = []

    statements = dict()
    cc = compile_starknet_files(files, cairo_path=cairo_path, debug_info=True)

    for pc in set(cc.program.debug_info.instruction_locations.keys()):
        instruct = cc.program.debug_info.instruction_locations[pc].inst
        file = process_file(instruct.input_file.filename)
        while True:
            # If file is auto generated discard it.
            if "autogen" not in file:
                lines = list(
                    range(
                        instruct.start_line,
                        instruct.end_line + 1,
                    )
                )
                if not file in statements:
                    statements[file] = set()
                statements[file].update(lines)
            if (
                instruct.parent_location is not None
            ):  # Continue until we have last parent location.
                instruct = instruct.parent_location[0]
                file = process_file(instruct.input_file.filename)
            else:
                break

    return statements



import os
from collections import defaultdict
from typing import Any, DefaultDict, Dict, List, Optional, Set

from starkware.cairo.lang.compiler.instruction import Instruction
from starkware.cairo.lang.compiler.program import ProgramBase
from starkware.cairo.lang.vm.builtin_runner import BuiltinRunner
from starkware.cairo.lang.vm.relocatable import MaybeRelocatable
from starkware.cairo.lang.vm.vm_core import RunContext, VirtualMachine

from nile_coverage import logger
from nile_coverage.common import CairoTraceReport
from nile_coverage.utils import process_file


def get_coverage_results():
    report_dict = OverrideVm.covered()
    statements = OverrideVm.statements()

    return CairoTraceReport(statements, report_dict)


class OverrideVm(VirtualMachine):
    def __init__(
        self,
        program: ProgramBase,
        run_context: RunContext,
        hint_locals: Dict[str, Any],
        static_locals: Optional[Dict[str, Any]] = None,
        builtin_runners: Optional[Dict[str, BuiltinRunner]] = None,
        program_base: Optional[MaybeRelocatable] = None,
    ):
        super().__init__(
            program=program,
            run_context=run_context,
            hint_locals=hint_locals,
            static_locals=static_locals,
            builtin_runners=builtin_runners,
            program_base=program_base,
        )

        self.old_end_run = (
            super().end_run
        )  # Save the old end run function to wrap it afterwards.
        self.old_run_instruction = (
            super().run_instruction
        )  # Save the old run instruction function to wrap it afterwards.
        self.old_as_vm_exception = (
            super().as_vm_exception
        )  # Save the old vm as exception function to wrap it afterwards.
        self.touched_pcs: List[int] = []

    def run_instruction(self, instruction: Instruction):
        """Saves the current pc and runs the instruction."""
        self.touched_pcs.append(self.run_context.pc.offset)
        self.old_run_instruction(instruction=instruction)

    def end_run(self):
        """In case the run doesn't fail creates report coverage."""
        self.old_end_run()
        self.cover_file()

    def as_vm_exception(
        self,
        exc,
        with_traceback: bool = True,
        notes: Optional[List[str]] = None,
        hint_index: Optional[int] = None,
    ):
        """In case the run fails creates report coverage."""
        self.cover_file()
        return self.old_as_vm_exception(exc, with_traceback, notes, hint_index)

    @staticmethod
    def covered(
        val: DefaultDict[str, Set[int]] = defaultdict(set)
    ) -> DefaultDict[str, set]:
        """To share the covered files between all the instances."""
        return val

    @staticmethod
    def statements(
        val: DefaultDict[str, Set[int]] = defaultdict(set)
    ) -> DefaultDict[str, set]:
        """To share the lines of codes in files between all the instances."""
        return val

    def pc_to_line(
        self,
        pc,
        statements: DefaultDict[str, set],
        report_dict: DefaultDict[str, Set[int]],
    ) -> None:
        """
        Converts the touched pcs to the line numbers
        of the original file and saves them.
        """
        should_update_report = (
            pc in self.touched_pcs
        )  # If the pc is not touched by the test don't report it.
        instruct = self.program.debug_info.instruction_locations[
            pc
        ].inst  # First instruction in the debug info.
        file = process_file(
            instruct.input_file.filename
        )  # Current analyzed file relative path.
        while True:
            if "autogen" not in file:  # If file is auto generated discard it.
                lines = list(
                    range(
                        instruct.start_line,
                        instruct.end_line + 1,
                    )
                )  # Get the lines touched.
                if should_update_report:
                    report_dict[file].update(lines)
                statements[file].update(lines)
            if (
                instruct.parent_location is not None
            ):  # Continue until we have last parent location.
                instruct = instruct.parent_location[0]
                file = process_file(instruct.input_file.filename)
            else:
                return

    def cover_file(
        self,
    ):
        """Add the coverage report in the report dict and all the lines of code."""
        if self.program.debug_info is not None:
            report_dict = self.__class__.covered()
            statements = self.__class__.statements()
            for pc in set(self.program.debug_info.instruction_locations.keys()):
                self.pc_to_line(pc=pc, report_dict=report_dict, statements=statements)



"""Coverage reporters."""

import json
import os
import os.path
import sys
import time
import xml.dom.minidom
from collections import defaultdict

from pycobertura import Cobertura
from pycobertura.reporters import TextReporter as CoberturaTextReporter

from nile_coverage import __url__, __version__, logger
from nile_coverage.common import COVERAGE_DIRECTORY, CoverageFile
from nile_coverage.utils import add_files_to_report, get_file_statements


class XmlReporter:
    """Cobertura-style XML reports."""

    def __init__(self, contracts_folder, statements, report_dict):
        """Initialize reporter."""
        self.statements = statements
        self.report_dict = report_dict
        self.contracts_folder = contracts_folder
        self.cairo_path = [contracts_folder]

        # add empty coverage files to report
        add_files_to_report(contracts_folder, self.report_dict)

        self.source_paths = set()
        self.packages = {}
        self.xml_out = None

    def report(self, outfile=None, as_string=False):
        """
        Generate a Cobertura-compatible XML report.

        `outfile` is a file object to write the XML to.
        """
        # Initial setup.
        outfile = outfile or sys.stdout

        # Create the DOM that will store the data.
        impl = xml.dom.minidom.getDOMImplementation()
        self.xml_out = impl.createDocument(None, "coverage", None)

        # Write header stuff.
        xcoverage = self.xml_out.documentElement
        xcoverage.setAttribute("version", __version__)
        xcoverage.setAttribute("timestamp", str(int(time.time() * 1000)))
        xcoverage.appendChild(
            self.xml_out.createComment(f" Generated by nile-coverage: {__url__} ")
        )

        # Call xml_file for each file in the data.
        for file, coverage in self.report_dict.items():
            if file.startswith(self.contracts_folder):
                self.xml_file(
                    CoverageFile(
                        statements=set(self.statements[file]),
                        covered=set(coverage),
                        name=file,
                    )
                )

        xsources = self.xml_out.createElement("sources")
        xcoverage.appendChild(xsources)

        # Populate the XML DOM with the source info.
        for path in sorted(self.source_paths):
            xsource = self.xml_out.createElement("source")
            xsources.appendChild(xsource)
            txt = self.xml_out.createTextNode(path)
            xsource.appendChild(txt)

        lnum_tot, lhits_tot = 0, 0
        bnum_tot, bhits_tot = 0, 0

        xpackages = self.xml_out.createElement("packages")
        xcoverage.appendChild(xpackages)

        # Populate the XML DOM with the package info.
        for pkg_name, pkg_data in sorted(self.packages.items()):
            class_elts, lhits, lnum, bhits, bnum = pkg_data
            xpackage = self.xml_out.createElement("package")
            xpackages.appendChild(xpackage)
            xclasses = self.xml_out.createElement("classes")
            xpackage.appendChild(xclasses)
            for _, class_elt in sorted(class_elts.items()):
                xclasses.appendChild(class_elt)
            xpackage.setAttribute("name", pkg_name.replace(os.sep, "."))
            xpackage.setAttribute("line-rate", "0")
            branch_rate = "0"
            xpackage.setAttribute("branch-rate", branch_rate)
            xpackage.setAttribute("complexity", "0")

            lnum_tot += lnum
            lhits_tot += lhits
            bnum_tot += bnum
            bhits_tot += bhits

        xcoverage.setAttribute("lines-valid", str(lnum_tot))
        xcoverage.setAttribute("lines-covered", str(lhits_tot))
        xcoverage.setAttribute("line-rate", rate(lhits_tot, lnum_tot))
        xcoverage.setAttribute("branches-covered", "0")
        xcoverage.setAttribute("branches-valid", "0")
        xcoverage.setAttribute("branch-rate", "0")
        xcoverage.setAttribute("complexity", "0")

        # Return the string without writing to file
        if as_string:
            return serialize_xml(self.xml_out)

        # Write the output file.
        cwd = os.getcwd()
        output = os.path.join(cwd, outfile)
        with open(output, "w") as f:
            f.write(serialize_xml(self.xml_out))

        # Return the total percentage.
        denom = lnum_tot + bnum_tot
        if denom == 0:
            pct = 0.0
        else:
            pct = 100.0 * (lhits_tot + bhits_tot) / denom
        return pct

    def xml_file(self, cf: CoverageFile):
        """Add to the XML report for a single file."""
        # Create the 'lines' and 'package' XML elements, which
        # are populated later.  Note that a package == a directory.
        cf.name = cf.name.replace("\\", "/")
        self.source_paths.add(cf.name)

        rel_name = cf.name

        if cf.nb_statements == 0:
            statements = get_file_statements([rel_name], self.cairo_path)
            if rel_name in statements:
                cf.statements = list(statements[rel_name])
                cf.nb_statements = len(cf.statements)

        dirname = os.path.dirname(rel_name) or "."
        dirname = "/".join(dirname.split("/")[:6])
        package_name = dirname.replace("/", ".")

        package = self.packages.setdefault(package_name, [{}, 0, 0, 0, 0])

        xclass = self.xml_out.createElement("class")

        xclass.appendChild(self.xml_out.createElement("methods"))

        xlines = self.xml_out.createElement("lines")
        xclass.appendChild(xlines)

        xclass.setAttribute("name", os.path.relpath(rel_name, dirname))
        xclass.setAttribute("filename", rel_name.replace("\\", "/"))
        xclass.setAttribute("complexity", "0")

        # For each statement, create an XML 'line' element.
        for line in sorted(cf.statements):
            xline = self.xml_out.createElement("line")
            xline.setAttribute("number", str(line))

            # Q: can we get info about the number of times a statement is
            # executed?  If so, that should be recorded here.
            if line in cf.covered:
                xline.setAttribute("hits", "1")
            else:
                xline.setAttribute("hits", "0")
            xlines.appendChild(xline)

        class_branches = 0.0
        class_br_hits = 0.0
        branch_rate = "0"
        line_rate = rate(cf.nb_covered, cf.nb_statements)

        # Finalize the statistics that are collected in the XML DOM.
        xclass.setAttribute("line-rate", line_rate)
        xclass.setAttribute("branch-rate", branch_rate)

        package[0][rel_name] = xclass
        package[1] += cf.nb_covered
        package[2] += cf.nb_statements
        package[3] += class_br_hits
        package[4] += class_branches


class TextReporter(XmlReporter):
    """CLI text reports."""

    def report(self):
        """Print the coverage summary of the project."""
        xml_string = XmlReporter.report(self, as_string=True)

        cobertura = Cobertura(xml_string)
        tr = CoberturaTextReporter(cobertura)

        logger.info(f"\n\n{tr.generate()}")


def serialize_xml(dom):
    """Serialize a minidom node to XML."""
    return dom.toprettyxml()


def rate(hit, num):
    """Return the fraction of `hit`/`num`, as a string."""
    if num == 0:
        return "1"
    else:
        return "%.4g" % (float(hit) / num)


def run_report(contracts_folder: str = "", xml: bool = False):
    logger.info("\nGenerating coverage report. This can take a minute...")

    if not os.path.isdir(contracts_folder):
        logger.info(
            f'\n\nNothing to report (couldn\'t find "{contracts_folder}" directory)'
        )

    # Aggregate nile.coverage files
    statements = defaultdict(set)
    report_dict = defaultdict(set)

    files = [
        os.path.join(COVERAGE_DIRECTORY, f)
        for f in os.listdir(COVERAGE_DIRECTORY)
        if os.path.isfile(os.path.join(COVERAGE_DIRECTORY, f))
    ]

    for f in files:
        with open(f, "r") as fp:
            data = json.load(fp)
            for contract in data["lines"]:
                statements[contract].update(data["lines"][contract])
            for contract in data["covered_lines"]:
                report_dict[contract].update(data["covered_lines"][contract])

    if xml:
        reporter = XmlReporter(contracts_folder, statements, report_dict)
        reporter.report(outfile="coverage.xml")
    else:
        reporter = TextReporter(contracts_folder, statements, report_dict)
        reporter.report()



"""xdist WorkerController overrides."""
import pytest

import nile_coverage.xdist.remote


class CustomRemoteHook:
    """RemoteHook extension to handle integration."""

    @pytest.hookimpl(trylast=True)
    def pytest_xdist_getremotemodule(self):
        return nile_coverage.xdist.remote



"""nile-coverage common module."""
import json
import shutil
from dataclasses import dataclass
from typing import DefaultDict, Set

from nile_coverage.utils import JsonEncoder

COVERAGE_DIRECTORY = "cairo-coverage"


@dataclass
class CairoTraceReport:
    """Wrapper for the coverage trace results."""

    lines: DefaultDict[str, set]
    covered_lines: DefaultDict[str, set]

    def __repr__(self):
        data = {
            "lines": self.lines,
            "covered_lines": self.covered_lines,
        }
        return json.dumps(data, indent=2, cls=JsonEncoder)


@dataclass
class CoverageFile:
    name: str  # Filename.
    covered: Set[int]  # Tested lines.
    statements: Set[int]  # Lines with code.

    def __post_init__(self):
        """Finish initialization."""
        # Number of lines with code in the cairo file.
        self.nb_statements = len(self.statements)
        # Number of lines tested.
        self.nb_covered = len(self.covered)


def clean():
    """Remove coverage files after execution."""
    shutil.rmtree(COVERAGE_DIRECTORY)



"""Integration plugins."""
import pytest
from xdist.workermanage import WorkerController

from nile_coverage.common import clean
from nile_coverage.vendor.reporters import run_report
from nile_coverage.xdist.worker import CustomRemoteHook


class PytestCairoCoveragePlugin:
    """Coverage Plugin."""

    def __init__(self, contracts_folder, xml=False):
        self.contracts_folder = contracts_folder
        self.xml = xml

    @pytest.hookimpl(hookwrapper=True)
    def pytest_sessionfinish(self):
        yield
        run_report(self.contracts_folder, self.xml)

        clean()


WorkerController.RemoteHook = CustomRemoteHook
